import cairo, rsvg, math
from const import *

# A Tile is a great user interface widget because it can display things,
# be interactive in its own right and is very descriptive visually and
# textually

# ASCII art
#-----------------------------------------------------------#
# +-----+ TITLE                              Button +-----+ #
# |  A  | Description                        Button +  B  + #
# +-----+ *****                              Status +-----+ #
#-----------------------------------------------------------#
# A Icon image
# B Animated image space
# some elements could be omitted, for instance buttons or icon
# or description can easily go. 

# Todo
#  underline
#  full justification
#  buttons
#  Animations
#  Status text

class Tile:
    def __init__( self, cr=None, render_now=False,  
                 title=None, 
                 icon=None, 
                 description=None, 
                 rating=None, 
                 buttons=[], 
                 rect=None ):
        if not cr:
            render_now = False
        else:
            self.cr = cr
        
        self.SetMaxButtons()
        self.buttons = buttons
        rect = (rect.x, rect.y, rect.width, rect.height)
        self.SetRect( rect )
        self.SetPadding()
        self.SetRating( rating )
        self.SetTitle( title )
        self.SetDescription( description )
        self.SetIcon( icon )
        
        if render_now:
            self.DrawTile()
        
    def SetTitle( self, title=None ):
        if not title:
            return
        self.title = title
        self.title_text_ratio = 0.225 # 18px(title_font)/80px(height)
        self.title_max_size = 18
        
        self.title_size = self.title_text_ratio * self.height
        if self.title_size > self.title_max_size:
            self.title_size = self.title_max_size
            
        if self.rating:
            self.title_y = self.title_size * 1.1 # FIXME
        else:
            self.title_y = self.title_size 

    def SetIcon( self, icon=None ):
        self.tile_info_x = self.height
        self.icon_height = self.height - ( 2 * self.padding )
        if not icon:
            return
        self.icon = icon
        
    def SetDescription( self, description=None ):
        if not description:
            return
        self.description = description
        self.description_height = self.height - (2*self.padding) - self.title_size - self.ratings_height
        self.description_ratio = 0.16
        self.description_max_size = 10
        self.description_size = self.description_ratio * self.height
        if self.description_size > self.description_max_size:
            self.description_size = self.description_max_size
        self.description_line_height = 1.2 * self.description_size
        self.description_rows = int(self.description_height / self.description_line_height)
        if self.rating:
            self.desc_y = self.title_y *1.7# FIXME    
        else:
            self.desc_y =  self.description_rows + self.description_size
        
    def SetRating( self, rating=None ):
        self.rating = rating
        
        if self.rating:
            self.ratings_ratio = 0.3
            self.ratings_height = self.ratings_ratio * self.height
        else:
            self.ratings_height = 0
        
    def SetRect( self, rect=None ):
        if not rect:
            rect = ( 0, 0, 500, 80 )
        ( self.x, self.y, self.width, self.height ) = rect
        
    def SetMaxButtons( self, buttons=4 ):
        self.max_buttons = buttons
        # Calculate text size for vertical buttons too
        
    def SetPadding( self, padding=5 ):
        self.padding = padding
    
    def GetRatingAtPosition(self, x, y):
        """
        Returns the rating clicked at a specific position
        """
        r_svg = rsvg.Handle( "../pixmaps/gold.svg" )
        tile_rating_y = self.icon_height - self.ratings_height + self.padding
        
        if y < tile_rating_y or y > tile_rating_y + self.ratings_height:
            return None
        
        self.ratings_width = (r_svg.props.width/r_svg.props.height)*self.ratings_height
        
        if x < self.tile_info_x or x > self.tile_info_x + (5 * self.ratings_width):
            return None

        return math.ceil(float(x - self.tile_info_x)/float(self.ratings_width))
    
    def GetButtonAtPosition(self, x, y):
        """
        Returns the button signal at a specific position
        """
        pass
    
    def AddButton( self, text, signal, properties, rect ):
        ( size, style, font, alignment, color ) = properties
        ( r, g, b ) = color
        if rect:
            ( x ,y, width, height ) = rect
        else:
            ( x ,y, width, height ) = ( None, None, None, None )
        self.buttons.append( x, y, signal, text, alignment, style, r, g, b )
        # Signal will probably be passed as data by the event handler
        
    def DrawTile( self ):
        # Move the canvas to the origin
        self.cr.translate( self.x, self.y )
        
        # Render each element separately
        self.DrawIcon()
        self.DrawTitle()
        self.DrawDescription()
        self.DrawRating()
        self.DrawButtons()
    
    def DrawIcon( self, x=None, y=None ):
        '''
        Simple image draw function, could equally be used for any kind of image
        image can be moved by x/y
        '''
        if not self.icon:
            return
        self.cr.save()
        if x and y:
            self.cr.translate( x, y )
        
        self.cr.translate( self.padding, self.padding )
        if self.icon.endswith( ".svg" ):
            # Render SVG files
            i_svg = rsvg.Handle( self.icon )
            self.icon_width = ( float( self.icon_height )/float( i_svg.props.height ) ) * float( i_svg.props.width )
            self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
            self.cr.scale( float( self.icon_width )/float( i_svg.props.width ), 
                           float( self.icon_height )/float( i_svg.props.height ) )
            i_svg.render_cairo( self.cr )
        elif self.icon.endswith( ".png" ):
            # Render PNG files     
            image = cairo.ImageSurface.create_from_png ( self.icon )
            p_width, p_height = image.get_width(), image.get_height()
            self.icon_width = ( float( self.icon_height )/float( p_height ) ) * float( p_width )
            self.cr.translate( ( self.icon_height - self.icon_width )/2, 0 )
            self.cr.scale( self.icon_width/float( p_width ), 
                           self.icon_width/float( p_height ))
            self.cr.set_source_surface( image, 0, 0 );
            self.cr.paint()
        self.cr.restore()
        
    def DrawTitle( self ):
        style = BOLD
        font = None
        effects = (None, None, CAIRO_HINT_STYLE_STROKE, None)
        alignment = LEFT
        color = (0,0,0)
        properties = ( self.title_size, style, font, alignment, color ) 
        self.DrawText(self.title, properties, (self.tile_info_x, self.title_y, None, None), effects)
        
    def DrawDescription( self ):
        style = NORMAL
        font = None
        effects = (None, None, CAIRO_HINT_STYLE_DEFAULT, None)
        alignment = LEFT
        color = (0.3,0.3,0.3)
        properties = ( self.description_size, style, font, alignment, color ) 
        self.DrawText(self.description, properties, (self.tile_info_x, self.desc_y, None, None))
        
    def DrawRating( self ):
        if not self.rating:
            return
        if self.rating:
            r_svg = rsvg.Handle( "../pixmaps/gold.svg" )
            
            tile_rating_y = self.icon_height - self.ratings_height + self.padding
            self.ratings_width = (r_svg.props.width/r_svg.props.height)*self.ratings_height
            
            self.cr.save()
            self.cr.translate( self.tile_info_x, tile_rating_y )
            self.cr.scale( float( self.ratings_width )/float( r_svg.props.width ), 
                           float( self.ratings_height )/float( r_svg.props.height ) )
            for x in range( 5 ):
                if x < self.rating:
                    r_svg.render_cairo( self.cr )
                else:
                    rating_radius = 3
                    self.cr.arc( r_svg.props.width/2, 
                           r_svg.props.height/2, 
                           rating_radius, 0, 2.0 * math.pi )
                    self.cr.set_source_rgb( 0.73, 0.73, 0.71 )
                    self.cr.fill_preserve()
                    self.cr.set_source_rgb( 0.56, 0.56, 0.53 )
                    self.cr.set_line_width( 1 )
                    self.cr.stroke()
                self.cr.translate( r_svg.props.width , 0 )
            self.cr.restore()
            
    def DrawButtons( self ):
        if len(self.buttons) == 0:
            return
        
    def DrawText( self, text, properties, rect=None, effects=None):
        ( size, style, font, alignment, color ) = properties 
        if effects:
            ( shadow, stroke, hinting, opacity ) = effects
            if shadow:
                ( shadow_offsetX, shadow_offsetY, shadow_opacity, shadow_size ) = shadow
            if stroke:
                ( stroke_width, stroke_color, stroke_opacity ) = stroke
            else:
                stroke_width = None
        else:
            ( shadow, stroke, hinting, opacity ) = (None, None, None, None)
            ( shadow_offsetX, shadow_offsetY, shadow_opacity, shadow_size ) = (None, None, None, None)
            ( stroke_width, stroke_color, stroke_opacity ) = (None, None, None)
        
        self.cr.save()
        if rect:
            ( x ,y, width, height ) = rect
        else:
            ( x ,y, width, height ) = ( 0, 0, None, None )
        
        self.cr.set_font_size( size )
        if not font:
            font = "Bitstream Vera Sans"
        # styles 
        if style & BOLD:
            weight = CAIRO_FONT_WEIGHT_BOLD
        else:
            weight = CAIRO_FONT_WEIGHT_NORMAL
            
        if style & ITALIC:
            slant = CAIRO_FONT_SLANT_ITALIC
        else:
            slant = CAIRO_FONT_SLANT_NORMAL
            
        self.cr.select_font_face( font, slant, weight)
        (fascent, fdescent, fheight, fxadvance, fyadvance) = self.cr.font_extents()

        # line splitting, word wrapping and ffr ellipsising
        
        # This should handle word wrapping correctly
        if width:
            lines = []
            line = ""
            text.replace("\n", "")
            for letter in text:
                line = line + letter
                (x_b, y_b, text_width, t_h, x_a, y_a) = self.cr.text_extents( line )
                if text_width > width:
                    line = line[0:len(line)-1]
                    lines.append(line)
                    line = letter
        else:
            lines = text.split("\n")
        
        if not y:
            y = 3 + fyadvance 
        
        # Calculate the real Y position from the supplied rectangle
        cursor_y = y - fyadvance
        
        # Work out which method we should use for hinting our fonts
        if not hinting:
            hinting =  CAIRO_HINT_STYLE_STROKE

        if stroke_width and hinting == CAIRO_HINT_STYLE_STROKE:
            hinting = CAIRO_HINT_STYLE_FULL
             
        #if style | UNDERLINE:
        #    uline_start = cursor_y*1.1
        #else:
        uline_start = None
            
        for index,line in enumerate(lines):
            (x_bearing, y_bearing, text_width, text_height, x_advance, y_advance) = self.cr.text_extents( text )
            # Justification
            cursor_x = self.tile_info_x
            if alignment == RIGHT:
                cursor_x = x + text_width
            elif alignment == CENTER:
                cursor_x = x + (self.tile_info_x/2)
            self.cr.save()
            self.cr.move_to(cursor_x, cursor_y)
            self.cr.text_path( line )
            ( r, g, b ) = color
            self.cr.set_source_rgb( r, g, b )
            if hinting == CAIRO_HINT_STYLE_STROKE:
                self.cr.fill_preserve()
                ( r, g, b ) = color
                self.cr.set_source_rgb( r, g, b )
                self.cr.set_line_width( float(size/200) )
                self.cr.stroke()
            elif hinting < CAIRO_HINT_STYLE_STROKE:
                font_opts = self.cr.get_font_options()
                font_opts.set_hint_style(CAIRO_HINT_NUM - index)
                if stroke_width:
                    self.cr.fill_preserve()
                    ( r, g, b ) = stroke_color
                    self.cr.set_source_rgba( r, g, b, stroke_opacity )
                    self.cr.set_line_width( stroke_width )
                    self.cr.stroke()
                else:
                    self.cr.fill()
            
            cursor_y = cursor_y + fheight
            self.cr.restore()
            self.cr.save()
            if uline_start:
                (x_bearing, y_bearing, line_width, line_height, x_advance, y_advance) = self.cr.text_extents( line )
                uline_end = uline_start + line_width
                ( r, g, b ) = color
                self.cr.set_source_rgb( r, g, b )
                self.cr.move_to( uline_start+self.tile_info_x, cursor_y*1.1 )
                self.cr.line_to( uline_end, cursor_y*1.1 )
                self.cr.set_line_width( size/20 )
                self.cr.stroke()
            self.cr.restore()
        self.cr.restore()            
